from __future__ import absolute_import, print_function, unicode_literals
import json
import unittest
from collections import OrderedDict

import six
from rest_framework import parsers

from drf_braces.parsers import SortedJSONParser, StrippingJSONParser


class TestSortedJSONParser(unittest.TestCase):
    def setUp(self):
        super(TestSortedJSONParser, self).setUp()
        self.parser = SortedJSONParser()

    def test_parser(self):
        content = json.dumps({'hello': 'world'}).encode('utf-8')
        stream = six.BytesIO(content)

        actual_data = self.parser.parse(stream=stream)

        self.assertEqual(actual_data, OrderedDict([('hello', 'world')]))

    def test_parser_invalid_json(self):
        content = (
            json.dumps({'hello': 'world'})
            .replace('"', "'")
            .encode('utf-8')
        )
        stream = six.BytesIO(content)

        with self.assertRaises(parsers.ParseError):
            self.parser.parse(stream=stream)


class TestStrippingJSONParser(unittest.TestCase):
    def setUp(self):
        super(TestStrippingJSONParser, self).setUp()
        self.parser = StrippingJSONParser()

    def test_parser(self):
        content = json.dumps({'root': {'hello': 'world'}}).encode('utf-8')
        stream = six.BytesIO(content)

        actual_data = self.parser.parse(
            stream=stream,
            parser_context={'parse_root': 'root'}
        )

        self.assertEqual(actual_data, OrderedDict([('hello', 'world')]))

    def test_parser_no_root(self):
        content = json.dumps({'root': {'hello': 'world'}}).encode('utf-8')
        stream = six.BytesIO(content)

        actual_data = self.parser.parse(
            stream=stream,
            parser_context={}
        )

        self.assertEqual(actual_data, {'root': {'hello': 'world'}})

    def test_parser_different_root(self):
        content = json.dumps({'root': {'hello': 'world'}}).encode('utf-8')
        stream = six.BytesIO(content)

        actual_data = self.parser.parse(
            stream=stream,
            parser_context={'parse_root': 'foo'}
        )

        self.assertEqual(actual_data, {'root': {'hello': 'world'}})



from __future__ import absolute_import, print_function, unicode_literals
import unittest

import mock
from rest_framework.generics import GenericAPIView

from drf_braces.mixins import (
    MapDataViewMixin,
    MultipleSerializersViewMixin,
    StrippingJSONViewMixin,
)


class TestMultipleSerializersViewMixin(unittest.TestCase):
    def setUp(self):
        super(TestMultipleSerializersViewMixin, self).setUp()

        class View(MultipleSerializersViewMixin, GenericAPIView):
            pass

        self.view = View()

    @mock.patch.object(GenericAPIView, 'get_serializer_context')
    @mock.patch.object(GenericAPIView, 'get_serializer_class')
    def test_get_serializer(self,
                            mock_get_serializer_class,
                            mock_get_serializer_context):
        context = {'context': 'here'}
        mock_get_serializer_context.return_value = context

        serializer = self.view.get_serializer(hello='world')

        self.assertEqual(serializer, mock_get_serializer_class.return_value.return_value)
        mock_get_serializer_class.assert_called_once_with()
        mock_get_serializer_class.return_value.assert_called_once_with(
            hello='world', context=context
        )
        mock_get_serializer_context.assert_called_once_with()

    @mock.patch.object(GenericAPIView, 'get_serializer_context')
    @mock.patch.object(GenericAPIView, 'get_serializer_class')
    def test_get_serializer_with_class(self,
                                       mock_get_serializer_class,
                                       mock_get_serializer_context):
        context = {'context': 'here'}
        mock_get_serializer_context.return_value = context
        serializer_class = mock.MagicMock()

        serializer = self.view.get_serializer(hello='world', serializer_class=serializer_class)

        self.assertEqual(serializer, serializer_class.return_value)
        self.assertFalse(mock_get_serializer_class.called)
        serializer_class.assert_called_once_with(hello='world', context=context)
        mock_get_serializer_context.assert_called_once_with()


class TestMapDataViewMixin(unittest.TestCase):
    def setUp(self):
        super(TestMapDataViewMixin, self).setUp()

        class View(MapDataViewMixin, GenericAPIView):
            pass

        self.view = View()
        self.view.request = mock.MagicMock(data=mock.sentinel.data)

    def test_get_data_no_mapper(self):
        actual = self.view.get_data()

        self.assertEqual(actual, mock.sentinel.data)

    @mock.patch.object(GenericAPIView, 'get_serializer_context')
    def test_get_data_attribute_mapper(self, mock_get_serializer_context):
        mapper = self.view.data_mapper_class = mock.MagicMock()
        actual = self.view.get_data()

        self.assertEqual(actual, mapper.return_value.return_value)
        mapper.assert_called_once_with(
            context=mock_get_serializer_context.return_value
        )
        mapper.return_value.assert_called_once_with(mock.sentinel.data)

    @mock.patch.object(GenericAPIView, 'get_serializer_context')
    def test_get_data_provided(self, mock_get_serializer_context):
        mapper = mock.MagicMock()
        actual = self.view.get_data(mapper_class=mapper)

        self.assertEqual(actual, mapper.return_value.return_value)
        mapper.assert_called_once_with(
            context=mock_get_serializer_context.return_value
        )
        mapper.return_value.assert_called_once_with(mock.sentinel.data)


class TestStrippingJSONViewMixin(unittest.TestCase):
    def setUp(self):
        super(TestStrippingJSONViewMixin, self).setUp()

        class View(StrippingJSONViewMixin, GenericAPIView):
            pass

        self.view = View()
        self.view.request = mock.MagicMock()

    def test_get_parser_context(self):
        self.view.parser_root = mock.sentinel.parser_root

        actual = self.view.get_parser_context(self.view.request)

        self.assertIn('parse_root', actual)
        self.assertEqual(actual['parse_root'], mock.sentinel.parser_root)



from __future__ import absolute_import, print_function, unicode_literals
import unittest
from collections import OrderedDict
from decimal import Decimal

import mock
import pytz

from drf_braces.fields.custom import (
    NonValidatingChoiceField,
    PositiveIntegerField,
    RoundedDecimalField,
    UTCDateTimeField,
    UnvalidatedField,
)


class TestUnvalidatedField(unittest.TestCase):
    def test_run_validators(self):
        validator = mock.MagicMock()
        field = UnvalidatedField(validators=[validator])

        actual = field.run_validators(mock.sentinel.value)

        self.assertIsNone(actual)
        self.assertFalse(validator.called)


class TestPositiveIntegerField(unittest.TestCase):
    def test_init_default(self):
        field = PositiveIntegerField()
        self.assertEqual(field.min_value, 0)

    def test_init_passed(self):
        field = PositiveIntegerField(min_value=mock.sentinel.min_value)
        self.assertEqual(field.min_value, mock.sentinel.min_value)


class TestUTCDateTimeField(unittest.TestCase):
    def test_init(self):
        field = UTCDateTimeField()

        self.assertEqual(
            getattr(field, 'timezone', getattr(field, 'default_timezone', None)),
            pytz.UTC
        )


class TestNonValidatingChoiceField(unittest.TestCase):
    def test_init(self):
        field = NonValidatingChoiceField()

        self.assertEqual(field.choices, OrderedDict())

    def test_to_internal_value(self):
        field = NonValidatingChoiceField(choices=['bar'])

        self.assertEqual(field.to_internal_value('bar'), 'bar')
        self.assertEqual(field.to_internal_value('haha'), 'haha')


class TestRoundedDecimalField(unittest.TestCase):
    def test_init(self):
        field = RoundedDecimalField()
        self.assertIsNotNone(field.max_digits)
        self.assertEqual(field.decimal_places, 2)

    def test_to_internal_value(self):
        field = RoundedDecimalField()
        self.assertEqual(field.to_internal_value(5), Decimal('5'))
        self.assertEqual(field.to_internal_value(5.2), Decimal('5.2'))
        self.assertEqual(field.to_internal_value(5.23), Decimal('5.23'))
        self.assertEqual(field.to_internal_value(5.2345), Decimal('5.23'))
        self.assertEqual(field.to_internal_value(5.2356), Decimal('5.24'))
        self.assertEqual(field.to_internal_value('5'), Decimal('5'))
        self.assertEqual(field.to_internal_value('5.2'), Decimal('5.2'))
        self.assertEqual(field.to_internal_value('5.23'), Decimal('5.23'))
        self.assertEqual(field.to_internal_value('5.2345'), Decimal('5.23'))
        self.assertEqual(field.to_internal_value('5.2356'), Decimal('5.24'))
        self.assertEqual(field.to_internal_value(Decimal('5')), Decimal('5'))
        self.assertEqual(field.to_internal_value(Decimal('5.2')), Decimal('5.2'))
        self.assertEqual(field.to_internal_value(Decimal('5.23')), Decimal('5.23'))
        self.assertEqual(field.to_internal_value(Decimal('5.2345')), Decimal('5.23'))
        self.assertEqual(field.to_internal_value(Decimal('5.2356')), Decimal('5.24'))
        self.assertEqual(field.to_internal_value(Decimal('4.2399')), Decimal('4.24'))



from __future__ import absolute_import, print_function, unicode_literals
import inspect

from rest_framework import fields, serializers
from rest_framework.fields import empty

from ..utils import add_base_class_to_instance, get_class_name_with_new_suffix


class EnforceValidationFieldMixin(object):
    """
    Custom DRF field mixin which allows to ignore validation error
    if the field is not mandatory.

    The field is mandatory when the parent serializer includes it
    in the ``must_validate_fields`` list or ``must_validate_fields``
    list is completely omitted. If the list is omitted, then all
    fields in that serializer are mandatory and must validate.
    """

    def run_validation(self, data=empty):
        try:
            return super(EnforceValidationFieldMixin, self).run_validation(data)
        except serializers.ValidationError:
            must_validate_fields = getattr(self.parent, 'must_validate_fields', None)
            field_name = getattr(self, 'field_name')

            # only re-raise validation error when this field must be validated
            # as defined by must_validate_fields list on the parent serializer
            # or if must_validate_fields is not defined
            if must_validate_fields is None or field_name in must_validate_fields:
                raise
            else:
                raise fields.SkipField(
                    'This field "{}" is being skipped as per enforce validation logic.'
                    ''.format(field_name)
                )


def _create_enforce_validation_serializer(serializer, strict_mode_by_default=True):
    """
    Recursively creates a copy of a given serializer which enforces ``must_validate_fields``.

    This function recursively copies serializers and replaces all fields
    by adding ``EnforceValidationFieldMixin`` to their mro which then enforces validation
    for fields defined in ``must_validate_fields`` and drops their data for all other fields.

    Args:
        serializer (Serializer): Serializer class or instance to be copied
        strict_mode_by_default (bool): Whether serializer should use strict mode
            when ``must_validate_fields`` is not defined.
            If ``True``, then all fields must be validated by default
            and if ``False``, then all fields can be dropped.

    Returns:
        Recursive copy of the ``serializer`` which will enforce ``must_validate_fields``.
    """
    if inspect.isclass(serializer):
        serializer = type(
            get_class_name_with_new_suffix(
                serializer,
                'Serializer',
                'EnforceValidationSerializer'
            ),
            (serializer,),
            {}
        )
        fields = serializer._declared_fields
        declared_fields = None

    else:
        # when serializer is instance, we still want to create a modified
        # serializer class copy since if we only adjust the fields
        # on the instance, that can complicate introspecting,
        # especially when ``_create_enforce_validation_serializer``
        # is used as decorator
        serializer = add_base_class_to_instance(
            serializer,
            new_name=get_class_name_with_new_suffix(
                serializer.__class__,
                'Serializer',
                'EnforceValidationSerializer'
            )
        )

        if isinstance(serializer, serializers.ListSerializer):
            serializer.child = _create_enforce_validation_serializer(
                serializer.child,
                strict_mode_by_default=strict_mode_by_default
            )

            # kwargs are used to take a deepcopy of the fields
            # so we need to adjust the child kwargs not to loose
            # reference to our custom child serializer
            if 'child' in serializer._kwargs:
                serializer._kwargs['child'] = serializer.child

            return serializer

        fields = serializer.fields
        declared_fields = serializer._declared_fields

    if not strict_mode_by_default and not hasattr(serializer, 'must_validate_fields'):
        serializer.must_validate_fields = []

        # this is necessary for deepcopy to work when
        # root serializer is instantiated it does deepcopy
        # on serializer._declared_fields which re-instantiates
        # all child fields hence must_validate_fields will be lost
        # adding it to the class makes it persistent
        if not inspect.isclass(serializer):
            serializer.__class__.must_validate_fields = []

    # cant use .items() since we need to adjust dictionary
    # within the loop so we cant be looping over the dict
    # at the same time
    # Python 3 even raises exception for this:
    # RuntimeError: dictionary changed size during iteration
    for name in list(fields.keys()):
        field = fields[name]
        replacement = None

        if isinstance(field, serializers.BaseSerializer):
            replacement = _create_enforce_validation_serializer(
                field,
                strict_mode_by_default=strict_mode_by_default
            )

        elif isinstance(field, serializers.Field):
            replacement = add_base_class_to_instance(
                field,
                EnforceValidationFieldMixin,
                new_name=get_class_name_with_new_suffix(
                    field.__class__,
                    'Field',
                    'EnforceValidationField'
                )
            )

        if replacement is not None:
            if declared_fields is not None:
                declared_fields[name] = replacement

            if replacement.source == name:
                replacement.source = None

            fields[name] = replacement

    return serializer


def create_enforce_validation_serializer(serializer=None, **kwargs):
    """
    Public function that creates a copy of a serializer which enforces ``must_validate_fields``.
    The difference between this function and ``_create_enforce_validation_serializer``
    is that this function can be used both as a direct decorator and decorator with
    parameters.

    For example::

        @create_enforce_validation_serializer
        class MySerializer(BaseSerializer): pass

        # or

        @create_enforce_validation_serializer(param=value)
        class MySerializer(BaseSerializer): pass

        # or

        create_enforce_validation_serializer(
            MySerializer,
            param=value
        )
    """
    # used as direct decorator so then simply return new serializer
    # e.g.  @decorator
    #       class MySerializer(...)
    # or used as regular function
    # e.g. function(Serializer, foo=bar)
    if inspect.isclass(serializer) and issubclass(serializer, serializers.Serializer):
        return _create_enforce_validation_serializer(serializer, **kwargs)

    # used as decorator with parameters
    # e.g.  @decorator(foo=bar)
    #       class MySerializer(...)
    elif serializer is None:
        def inner(serializer):
            return _create_enforce_validation_serializer(serializer, **kwargs)

        return inner

    else:
        raise TypeError(
            'create_enforce_validation_serializer can only be only on serializers. '
            'It was called with "{}"'.format(type(serializer))
        )



from __future__ import absolute_import, print_function, unicode_literals
from collections import OrderedDict

import six
from django import forms
from rest_framework import serializers

from .. import fields
from ..utils import (
    find_matching_class_kwargs,
    get_attr_from_base_classes,
    get_class_name_with_new_suffix,
    reduce_attr_dict_from_instance,
)


class FormSerializerFailure(object):
    """
    Enum for the possible form validation failure modes.

    'fail': validation failures should be added to self.errors
        and `is_valid()` should return False.

    'drop': validation failures for a given attribute will result in
        that attribute being dropped from `cleaned_data`;
        `is_valid()` will return True.

    'ignore': validation failures will be ignored, and the (invalid)
        data provided will be preserved in `cleaned_data`.
    """
    fail = 'fail'
    drop = 'drop'
    ignore = 'ignore'


class FormSerializerFieldMixin(object):
    def run_validation(self, data):
        try:
            return super(FormSerializerFieldMixin, self).run_validation(data)
        except (serializers.ValidationError, forms.ValidationError):
            # Only handle a ValidationError if the full validation is
            # requested or if field is in minimum required in the case
            # of partial validation.
            if any([not self.parent.partial,
                    self.parent.Meta.failure_mode == FormSerializerFailure.fail,
                    self.field_name in self.parent.Meta.minimum_required]):
                raise
            raise serializers.SkipField


def make_form_serializer_field(field_class):
    return type(
        get_class_name_with_new_suffix(field_class, 'Field', 'FormSerializerField'),
        (FormSerializerFieldMixin, field_class,),
        {}
    )


FORM_SERIALIZER_FIELD_MAPPING = {
    forms.CharField: make_form_serializer_field(fields.CharField),
    forms.MultipleChoiceField: make_form_serializer_field(fields.ChoiceField),
    forms.ChoiceField: make_form_serializer_field(fields.ChoiceField),
    forms.BooleanField: make_form_serializer_field(fields.BooleanField),
    forms.IntegerField: make_form_serializer_field(fields.IntegerField),
    forms.EmailField: make_form_serializer_field(fields.EmailField),
    forms.DateTimeField: make_form_serializer_field(fields.DateTimeField),
    forms.DateField: make_form_serializer_field(fields.DateField),
    forms.TimeField: make_form_serializer_field(fields.TimeField),
}


class FormSerializerOptions(object):
    """
    Defines what options FormSerializer can have in Meta.

    :param form: The ``django.form.Form`` class to use as the base
        for the serializer.
    :param failure_mode: `FormSerializerFailure`
    :param minimum_required: the minimum required fields that
        must validate in order for validation to succeed.
    """

    def __init__(self, meta, class_name):
        self.form = getattr(meta, 'form', None)
        self.failure_mode = getattr(meta, 'failure_mode', FormSerializerFailure.fail)
        self.minimum_required = getattr(meta, 'minimum_required', [])
        self.field_mapping = getattr(meta, 'field_mapping', {})

        assert self.form, (
            'Class {serializer_class} missing "Meta.form" attribute'.format(
                serializer_class=class_name
            )
        )
        assert self.failure_mode in vars(FormSerializerFailure).values(), (
            'Failure mode "{}" is not supported'.format(self.failure_mode)
        )
        if self.failure_mode == FormSerializerFailure.ignore:
            raise NotImplementedError(
                'Failure mode "{}" is not supported since it is not clear '
                'what is an expected behavior'.format(self.failure_mode)
            )

        # copy all other custom keys
        for k, v in vars(meta).items():
            if hasattr(self, k):
                continue
            setattr(self, k, v)


class FormSerializerMeta(serializers.SerializerMetaclass):
    def __new__(cls, name, bases, attrs):
        try:
            parents = [b for b in bases if issubclass(b, FormSerializer)]
        except NameError:
            # We are defining FormSerializer itself
            parents = None

        if not parents or attrs.pop('_is_base', False):
            return super(FormSerializerMeta, cls).__new__(cls, name, bases, attrs)

        assert 'Meta' in attrs, (
            'Class {serializer_class} missing "Meta" attribute'.format(
                serializer_class=name
            )
        )
        options_class = get_attr_from_base_classes(
            bases, attrs, '_options_class', default=FormSerializerOptions
        )
        attrs['Meta'] = options_class(attrs['Meta'], name)

        return super(FormSerializerMeta, cls).__new__(cls, name, bases, attrs)


class FormSerializerBase(serializers.Serializer):
    """
    The base Form serializer class.
    When a subclassing serializer is validated or saved, this will
    pass-through those operations to the mapped Form.
    """
    _is_base = True
    _options_class = FormSerializerOptions

    def __init__(self, *args, **kwargs):
        # We override partial validation handling, since for
        # it to be properly implemented for a Form the caller
        # must also choose whether or not to include the data
        # that failed validation in the result cleaned_data.
        # Unfortunately there is no way to prevent a caller from
        # sending this param themselves, because of the way DRFv2
        # serializers work internally.
        if self.Meta.failure_mode != FormSerializerFailure.fail:
            kwargs['partial'] = True

        self.form_instance = None

        super(FormSerializerBase, self).__init__(*args, **kwargs)

    def get_form(self, data=None, **kwargs):
        """
        Create an instance of configured form class.

        :param data: optional initial data
        :param kwargs: key args to pass to form instance
        :return: instance of `self.opts.form`, bound if data was provided,
            otherwise unbound.
        """
        form_cls = self.Meta.form

        instance = form_cls(data=data, **kwargs)

        # Handle partial validation on the form side
        if self.partial:
            set_form_partial_validation(
                instance, self.Meta.minimum_required
            )

        return instance

    def get_fields(self):
        """
        Return all the fields that should be serialized for the form.
        This is a hook provided by parent class.
        :return: dict of {'field_name': serializer_field_instance}
        """
        ret = super(FormSerializerBase, self).get_fields()

        field_mapping = reduce_attr_dict_from_instance(
            self,
            lambda i: getattr(getattr(i, 'Meta', None), 'field_mapping', {}),
            FORM_SERIALIZER_FIELD_MAPPING
        )

        # Iterate over the form fields, creating an
        # instance of serializer field for each.
        for field_name, form_field in self.Meta.form.base_fields.items():
            # if field is already defined via declared fields
            # skip mapping it from forms which then honors
            # the custom validation defined on the DRF declared field
            if field_name in ret:
                continue

            try:
                serializer_field_class = field_mapping[form_field.__class__]
            except KeyError:
                raise TypeError(
                    "{field} is not mapped to a serializer field. "
                    "Please add {field} to {serializer}.Meta.field_mapping. "
                    "Currently mapped fields: {mapped}".format(
                        field=form_field.__class__.__name__,
                        serializer=self.__class__.__name__,
                        mapped=', '.join(sorted([i.__name__ for i in field_mapping.keys()]))
                    )
                )
            else:
                ret[field_name] = self._get_field(form_field, serializer_field_class)

        return ret

    def _get_field(self, form_field, serializer_field_class):
        kwargs = self._get_field_kwargs(form_field, serializer_field_class)

        field = serializer_field_class(**kwargs)

        for kwarg, value in kwargs.items():
            # set corresponding DRF attributes which don't have alternative
            # in Django form fields
            if kwarg == 'required':
                field.allow_blank = not value
                field.allow_null = not value

            # ChoiceField natively uses choice_strings_to_values
            # in the to_internal_value flow
            elif kwarg == 'choices':
                field.choice_strings_to_values = {
                    six.text_type(key): key for key in OrderedDict(value).keys()
                }

        return field

    def _get_field_kwargs(self, form_field, serializer_field_class):
        """
        For a given Form field, determine what validation attributes
        have been set.  Includes things like max_length, required, etc.
        These will be used to create an instance of ``rest_framework.fields.Field``.

        :param form_field: a ``django.forms.field.Field`` instance
        :return: dictionary of attributes to set
        """
        attrs = find_matching_class_kwargs(form_field, serializer_field_class)

        if getattr(form_field, 'initial', None):
            attrs['default'] = form_field.initial

        # avoid "May not set both `required` and `default`"
        if attrs.get('required') and 'default' in attrs:
            del attrs['required']

        return attrs

    def validate(self, data):
        """
        Validate a form instance using the data that has been run through
        the serializer field validation.

        :param data: deserialized data to validate
        :return: validated, cleaned form data
        :raise: ``django.core.exceptions.ValidationError`` on failed
            validation.
        """
        self.form_instance = form = self.get_form(data=data)

        if not form.is_valid():
            _cleaned_data = getattr(form, 'cleaned_data', None) or {}

            if self.Meta.failure_mode == FormSerializerFailure.fail:
                raise serializers.ValidationError(form.errors)

            else:
                cleaned_data = data
                # use any cleaned data form might of validated right until
                # this moment even if validation failed
                cleaned_data.update(_cleaned_data)

        else:
            cleaned_data = form.cleaned_data

        return cleaned_data

    def to_representation(self, instance):
        """
        It doesn't make much sense to serialize a Form instance to JSON.
        """
        raise NotImplementedError(
            '{} does not currently serialize Form --> JSON'
            ''.format(self.__class__.__name__)
        )


class FormSerializer(six.with_metaclass(FormSerializerMeta, FormSerializerBase)):
    pass


class LazyLoadingValidationsMixin(object):
    """
    Provides a method for re-evaluating the validations for
    a form using an instance of it (whereas the FormSerializer
    only uses the form class).
    If your form class loads validations in `__init__()`, you
    need this.
    """

    def repopulate_form_fields(self):
        """
        Repopulate the form fields, update choices.
        The repopulation is required b/c some DT forms use a lazy-load approach
        to populating choices of a ChoiceField, by putting the load
        in the form's constructor.  Also, the DT fields may require context_data,
        which is unavailable when the fields are first constructed
        (which happens during evaluation of the serializer classes).
        :return: None
        """
        instance = self.get_form()

        for form_field_name, form_field in instance.fields.items():
            if hasattr(form_field, 'choices'):
                self.fields[form_field_name].choices = OrderedDict(form_field.choices)
                self.fields[form_field_name].choice_strings_to_values = {
                    six.text_type(key): key for key in OrderedDict(form_field.choices).keys()
                }

    def to_internal_value(self, data):
        """
        We have tons of "choices" loading in form `__init__()`,
        (so that DB query is evaluated at last possible moment) so require the
        use of ``common.common_json.serializers.LazyLoadingValidationsMixin``.
        """
        self.repopulate_form_fields()
        return super(LazyLoadingValidationsMixin, self).to_internal_value(data)


def set_form_partial_validation(form, minimum_required):
    """
    Get a form ready for partial validation.
    For all fields not in `minimum_required`, set
    `Field.required` to False.

    :param minimum_required: list of minimum required fields
    :return: None
    """
    for field_name, field in form.fields.items():
        if field_name not in minimum_required:
            field.required = False



from __future__ import absolute_import, print_function, unicode_literals

from .parsers import StrippingJSONParser


class MultipleSerializersViewMixin(object):
    def get_serializer(self, *args, **kwargs):
        serializer_class = kwargs.pop('serializer_class', None)
        if serializer_class is None:
            serializer_class = self.get_serializer_class()

        kwargs['context'] = self.get_serializer_context()

        return serializer_class(*args, **kwargs)


class MapDataViewMixin(object):
    # Configuration for data mapper.
    # Leave None if you don't require mapping
    data_mapper_class = None

    def get_mapper_context(self):
        return self.get_serializer_context()

    def get_data(self, mapper_class=None):
        """
        Get data for serialization.
        """
        if not mapper_class:
            mapper_class = self.data_mapper_class

        if mapper_class is not None:
            return mapper_class(context=self.get_mapper_context())(self.request.data)

        return self.request.data


class StrippingJSONViewMixin(object):
    parser_classes = (StrippingJSONParser,)

    # Location of the root of the data for this api.
    parser_root = None

    def get_parser_context(self, http_request):
        """
        Add 'parser_root' to this view's parser's parse_context.
        Since this view uses a DescendingJSONParser, it uses
        this information to decide what to pull out.
        """
        context = super(StrippingJSONViewMixin, self).get_parser_context(http_request)
        context.update({'parse_root': self.parser_root})
        return context



from __future__ import absolute_import, print_function, unicode_literals
import unittest

from drf_braces.fields import _fields
from drf_braces.fields.mixins import (
    AllowBlankNullFieldMixin,
    EmptyStringFieldMixin,
)


class TestFields(unittest.TestCase):
    def test_subclasses(self):
        for f in _fields.FIELDS:
            f = getattr(_fields, f)
            self.assertTrue(issubclass(f, EmptyStringFieldMixin))
            self.assertTrue(issubclass(f, AllowBlankNullFieldMixin))



from __future__ import absolute_import, print_function, unicode_literals
import inspect

from . import _fields as fields


class BooleanField(fields.BooleanField):
    def __init__(self, *args, **kwargs):
        self.TRUE_VALUES = self.TRUE_VALUES | set(kwargs.pop('true_values', []))
        self.FALSE_VALUES = self.FALSE_VALUES | set(kwargs.pop('false_values', []))
        super(BooleanField, self).__init__(*args, **kwargs)


class DecimalField(fields.DecimalField):
    def __init__(self, max_digits=None, decimal_places=None, *args, **kwargs):
        super(DecimalField, self).__init__(
            max_digits=max_digits, decimal_places=decimal_places,
            *args, **kwargs
        )

    def quantize(self, value):
        if self.max_digits is None or self.decimal_places is None:
            return value
        return super(DecimalField, self).quantize(value)


class DateTimeField(fields.DateTimeField):
    def __init__(self, *args, **kwargs):
        super(DateTimeField, self).__init__(*args, **kwargs)
        # allow fo the case when default_timezone is passed as None
        # since super will in that use Django's default timezone
        if 'default_timezone' in kwargs:
            # DRF >= 3.3
            if callable(self.default_timezone):
                self.timezone = kwargs['default_timezone']
            else:
                self.default_timezone = kwargs['default_timezone']


__all__ = [name for name, value in locals().items()
           if inspect.isclass(value) and issubclass(value, fields.Field)]



from __future__ import absolute_import, print_function, unicode_literals
import unittest
from datetime import datetime

import pytz
from dateutil.tz import tzoffset
from django import forms
from rest_framework import ISO_8601

from drf_braces.forms.fields import ISO8601DateTimeField


class TestISO8601DateTimeField(unittest.TestCase):
    def test_to_python_empty(self):
        field = ISO8601DateTimeField(required=False)

        self.assertIsNone(field.clean(''))

    def test_to_python_not_iso8601(self):
        field = ISO8601DateTimeField()

        self.assertEqual(field.clean('2015-01-01 16:30'), datetime(2015, 1, 1, 16, 30))
        with self.assertRaises(forms.ValidationError):
            field.clean('2015-01-01T16:30')

    def test_to_python_iso8601(self):
        field = ISO8601DateTimeField(input_formats=[ISO_8601])

        self.assertEqual(
            field.clean('2015-01-01 16:30'),
            datetime(2015, 1, 1, 16, 30)
        )
        self.assertEqual(
            field.clean('2015-01-01T16:30'),
            datetime(2015, 1, 1, 16, 30)
        )
        self.assertEqual(
            field.clean('2015-01-01T16:30+00:00'),
            datetime(2015, 1, 1, 16, 30).replace(tzinfo=pytz.UTC)
        )
        self.assertEqual(
            field.clean('2015-01-01T16:30+04:00'),
            datetime(2015, 1, 1, 16, 30).replace(tzinfo=tzoffset(None, 4 * 60 * 60))
        )
        with self.assertRaises(forms.ValidationError):
            field.clean('2015-01-01T16:30+A')



from __future__ import absolute_import, print_function, unicode_literals
import unittest
from decimal import Decimal

import pytz
from django.test.utils import override_settings

from drf_braces.fields.modified import BooleanField, DateTimeField, DecimalField


class TestBooleanField(unittest.TestCase):
    def test_init(self):
        field = BooleanField(true_values=['Y', 'Yes'], false_values=['N', 'No'])
        self.assertIn('Y', field.TRUE_VALUES)
        self.assertIn('Yes', field.TRUE_VALUES)
        self.assertIn('N', field.FALSE_VALUES)
        self.assertIn('No', field.FALSE_VALUES)


class TestDecimalField(unittest.TestCase):
    def test_init(self):
        field = DecimalField()
        self.assertIsNone(field.max_digits)
        self.assertIsNone(field.decimal_places)

    def test_quantize(self):
        field = DecimalField()
        self.assertIsNone(field.quantize(None))

        field = DecimalField(max_digits=4, decimal_places=3)
        self.assertEqual(field.quantize(Decimal('5.1234567')), Decimal('5.123'))


class TestDateTimeField(unittest.TestCase):
    @override_settings(USE_TZ=True)
    def test_init(self):
        value = '2015-01-02T16:00'

        self.assertIsNotNone(DateTimeField(default_timezone=pytz.utc).run_validation(value).tzinfo)
        self.assertIsNone(DateTimeField(default_timezone=None).run_validation(value).tzinfo)



from __future__ import absolute_import, print_function, unicode_literals
import unittest
from datetime import datetime

import mock
import six
from django import forms
from rest_framework import serializers

from drf_braces.forms.serializer_form import (
    SERIALIZER_FORM_FIELD_MAPPING,
    SerializerForm,
    SerializerFormBase,
    SerializerFormMeta,
    SerializerFormOptions,
    form_from_serializer,
)


TESTING_MODULE = 'drf_braces.forms.serializer_form'


class TestSerializer(serializers.Serializer):
    foo = serializers.CharField(read_only=True)
    bar = serializers.IntegerField(min_value=0)
    exclude = serializers.DateTimeField()

    def validate_bar(self, value):
        if value >= 1000:
            raise serializers.ValidationError('Serializer failure')
        return value


class TestForm(SerializerForm):
    class Meta(object):
        serializer = TestSerializer
        fields = ['bar', 'foo']
        exclude = ['exclude']


class TestSerializerFormOptions(unittest.TestCase):
    def test_init(self):
        _options = mock.Mock(serializer=serializers.Serializer)

        options = SerializerFormOptions(_options)

        self.assertEqual(options.serializer, _options.serializer)
        self.assertEqual(options.fields, _options.fields)
        self.assertEqual(options.exclude, _options.exclude)
        self.assertEqual(options.field_mapping, _options.field_mapping)

    def test_init_no_serializer(self):
        with self.assertRaises(AssertionError):
            SerializerFormOptions(None)

    def test_init_not_drf_serializer(self):
        with self.assertRaises(AssertionError):
            SerializerFormOptions(mock.Mock(serializer=int))


class TestSerializerFormMeta(unittest.TestCase):
    @mock.patch.dict(SERIALIZER_FORM_FIELD_MAPPING, {'base': 'mapping'}, clear=True)
    def test_get_field_mapping(self):
        class Base1B(object):
            class _meta(object):
                field_mapping = {
                    'hello': 'mars',
                    'hi': 'there',
                    'foo': 'bar',
                }

        class Base1A(Base1B):
            class _meta(object):
                field_mapping = {
                    'hello': 'world',
                }

        class Base2(object):
            class _meta(object):
                field_mapping = {
                    'hello': 'sun',
                    'something': 'here',
                }

        class Options(object):
            field_mapping = {
                'foo': 'baar',
                'happy': 'rainbows',
            }

        mapping = SerializerFormMeta.get_field_mapping((Base1A, Base2), Options)

        self.assertDictEqual(mapping, {
            'hi': 'there',
            'hello': 'world',
            'something': 'here',
            'base': 'mapping',
            'foo': 'baar',
            'happy': 'rainbows',
        })

    def test_get_form_fields_from_serializer(self):
        fields = SerializerFormMeta.get_form_fields_from_serializer(
            (object,), mock.Mock(serializer=TestSerializer,
                                 fields=['foo', 'bar'],
                                 exclude=['exclude'],
                                 field_mapping={})
        )

        self.assertSetEqual(set(fields.keys()), {'bar'})
        self.assertIsInstance(fields['bar'], forms.IntegerField)

    def test_get_form_fields_from_serializer_unmapped_field(self):
        class CustomField(serializers.Field):
            pass

        class MySerializer(serializers.Serializer):
            foo = serializers.CharField(read_only=True)
            bar = CustomField()

        with self.assertRaises(TypeError):
            SerializerFormMeta.get_form_fields_from_serializer(
                (object,), mock.Mock(serializer=MySerializer,
                                     fields=['foo', 'bar'],
                                     exclude=[],
                                     field_mapping={})
            )

    @mock.patch(TESTING_MODULE + '.SerializerFormOptions')
    def test_new_no_parents(self, mock_serializer_form_options):
        class TestForm(six.with_metaclass(SerializerFormMeta, forms.Form)):
            pass

        self.assertFalse(mock_serializer_form_options.called)

    def test_new(self):
        class TestForm(six.with_metaclass(SerializerFormMeta, SerializerForm)):
            class Meta(object):
                serializer = TestSerializer
                fields = ['bar', 'foo']
                exclude = ['exclude']

        self.assertIsInstance(TestForm._meta, SerializerFormOptions)
        self.assertSetEqual(set(TestForm.base_fields.keys()), {'bar'})


class TestSerializerFormBase(unittest.TestCase):
    def test_init(self):
        self.assertIsNone(SerializerFormBase().serializer)

    def test_get_serializer_context(self):
        self.assertDictEqual(SerializerFormBase().get_serializer_context(), {})

    def test_get_serializer_data(self):
        form = SerializerFormBase()
        form.cleaned_data = {'hello': 'world'}
        form.initial = {'stuff': 'here', 'hello': 'mars'}

        self.assertDictEqual(form.get_serializer_data(), {
            'hello': 'world',
            'stuff': 'here',
        })

    @mock.patch.object(SerializerFormBase, 'get_serializer_data')
    @mock.patch.object(SerializerFormBase, 'get_serializer_context')
    def test_get_serializer(self, mock_get_serializer_context, mock_get_serializer_data):
        form = SerializerFormBase()
        mock_serializer = mock.Mock()
        form._meta = mock.Mock(serializer=mock_serializer)

        actual = form.get_serializer()

        self.assertEqual(actual, mock_serializer.return_value)
        mock_serializer.assert_called_once_with(
            data=mock_get_serializer_data.return_value,
            context=mock_get_serializer_context.return_value,
        )

    @mock.patch.object(SerializerFormBase, 'get_serializer')
    def test_clean_form_valid(self, mock_get_serializer):
        class TestForm(SerializerFormBase):
            hello = forms.CharField()

        form = TestForm(data={'hello': 'world'})
        mock_get_serializer.return_value.is_valid.return_value = True
        mock_get_serializer.return_value.validated_data = {
            'and': 'mars'
        }

        self.assertTrue(form.is_valid(), form.errors)
        self.assertEqual(form.cleaned_data, {
            'hello': 'world',
            'and': 'mars',
        })

    @mock.patch.object(SerializerFormBase, 'get_serializer')
    def test_clean_form_invalid(self, mock_get_serializer):
        form = SerializerFormBase(data={})
        mock_get_serializer.return_value.is_valid.return_value = False
        mock_get_serializer.return_value.errors = {
            'field': ['error'],
        }

        self.assertFalse(form.is_valid(), form.errors)

        self.assertEqual(form.errors, {'field': ['error']})


class TestSerializerForm(unittest.TestCase):
    def test_full_clean_valid(self):
        data = {
            'foo': 'anything',
            'bar': '500',
            'exclude': '2015-01-01T12:00'
        }
        initial = {
            'exclude': datetime(2016, 1, 1, 16, 30)
        }

        form = TestForm(data=data, initial=initial)

        self.assertTrue(form.is_valid(), dict(form.errors))
        self.assertDictEqual(form.cleaned_data, {
            'bar': 500,
            'exclude': datetime(2016, 1, 1, 16, 30)
        })

    def test_full_clean_invalid(self):
        data = {
            'foo': 'anything',
            'bar': '1000',
            'exclude': '2015-01-01T12:00'
        }
        initial = {
            'exclude': datetime(2016, 1, 1, 16, 30)
        }

        form = TestForm(data=data, initial=initial)

        self.assertFalse(form.is_valid())
        self.assertIn('bar', form.errors)


class TestUtils(unittest.TestCase):
    def test_form_from_serializer_not_serializer(self):
        with self.assertRaises(AssertionError):
            form_from_serializer(None)

    def test_form_from_serializer(self):
        form = form_from_serializer(TestSerializer, fields=['foo', 'bar'], exclude=['exclude'])

        self.assertTrue(issubclass(form, SerializerForm))
        self.assertIsInstance(form._meta, SerializerFormOptions)
        self.assertListEqual(form._meta.fields, ['foo', 'bar'])
        self.assertListEqual(form._meta.exclude, ['exclude'])
        self.assertIs(form._meta.serializer, TestSerializer)



from __future__ import absolute_import, print_function, unicode_literals
import inspect

from rest_framework import fields
from rest_framework.fields import *  # noqa
from rest_framework.fields import _UnvalidatedField  # noqa

from .mixins import AllowBlankNullFieldMixin, EmptyStringFieldMixin


FIELDS = [
    'BooleanField',
    'CharField',
    'ChoiceField',
    'DateField',
    'DateTimeField',
    'DecimalField',
    'DurationField',
    'EmailField',
    'FileField',
    'FloatField',
    'HiddenField',
    'ImageField',
    'IntegerField',
    'IPAddressField',
    'MultipleChoiceField',
    'NullBooleanField',
    'RegexField',
    'SlugField',
    'TimeField',
    'URLField',
    'UUIDField',
]


def get_updated_fields(fields, base_classes):
    fields = [globals()[i] for i in fields]
    return {
        field.__name__: type(field.__name__, base_classes + (field,), {})
        for field in fields
    }


locals().update(
    get_updated_fields(FIELDS, (EmptyStringFieldMixin, AllowBlankNullFieldMixin))
)

__all__ = [name for name, value in locals().items()
           if inspect.isclass(value) and issubclass(value, fields.Field)]



from __future__ import absolute_import, print_function, unicode_literals
import unittest

from rest_framework import fields

from drf_braces.utils import (
    find_class_args,
    find_function_args,
    get_attr_from_base_classes,
    get_class_name_with_new_suffix,
)


class TestUtils(unittest.TestCase):
    def test_get_class_name_with_new_suffix(self):
        new_name = get_class_name_with_new_suffix(
            klass=fields.IntegerField,
            existing_suffix='Field',
            new_suffix='StrawberryFields'
        )
        self.assertEqual(new_name, 'IntegerStrawberryFields')

        new_name = get_class_name_with_new_suffix(
            klass=fields.IntegerField,
            existing_suffix='straws',
            new_suffix='Blueberries'
        )
        self.assertEqual(new_name, 'IntegerFieldBlueberries')

    def test_get_attr_from_base_classes(self):
        Parent = type(str('Parent'), (), {'fields': 'pancakes'})

        self.assertEqual(
            get_attr_from_base_classes((Parent,), [], 'fields'), 'pancakes'
        )

        self.assertEqual(
            get_attr_from_base_classes(
                (Parent,), {'fields': 'mushrooms'}, 'fields'
            ),
            'mushrooms'
        )

        self.assertEqual(
            get_attr_from_base_classes((Parent,), [], '', default='maple_syrup'),
            'maple_syrup'
        )

        with self.assertRaises(AttributeError):
            get_attr_from_base_classes(
                (Parent,), {'fields': 'mushrooms'}, 'catchmeifyoucan'
            )

    def test_find_function_args(self):
        def foo(a, b, c):
            pass

        self.assertListEqual(find_function_args(foo), ['a', 'b', 'c'])

    def test_find_function_args_invalid(self):
        self.assertListEqual(find_function_args(None), [])

    def test_find_class_args(self):
        class Bar(object):
            def __init__(self, a, b):
                pass

        class Foo(Bar):
            def __init__(self, c, d):
                super(Foo, self).__init__(None, None)
                pass

        self.assertSetEqual(set(find_class_args(Foo)), {'a', 'b', 'c', 'd'})



from __future__ import absolute_import, print_function, unicode_literals
import inspect

import pytz
import six
from django.utils.translation import gettext as _

from . import _fields as fields
from .mixins import ValueAsTextFieldMixin


class UnvalidatedField(fields._UnvalidatedField):
    """
    Same as DRF's ``_UnvalidatedField``, except this is a public class.
    """

    def run_validators(self, value):
        return


class PositiveIntegerField(fields.IntegerField):
    """
    Enhanced DRF's ``IntegerField`` as this default ``min_value`` to be 0
    hence only allowing positive numbers.
    """

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('min_value', 0)
        super(PositiveIntegerField, self).__init__(*args, **kwargs)


class UTCDateTimeField(fields.DateTimeField):
    """
    Same as DateTimeField except this field guarantees to return time-zone aware dates.
    """

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('default_timezone', pytz.utc)
        super(UTCDateTimeField, self).__init__(*args, **kwargs)


class NonValidatingChoiceField(fields.ChoiceField):
    """
    ChoiceField subclass that skips the validation of "choices".
    It does apply 'required' validation, and any other validation
    done by the parent drf.Field class.
    """

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('choices', [])
        super(NonValidatingChoiceField, self).__init__(*args, **kwargs)

    def to_internal_value(self, data):
        try:
            return self.choice_strings_to_values[six.text_type(data)]
        except KeyError:
            return six.text_type(data)


class NumericField(ValueAsTextFieldMixin, fields.IntegerField):
    default_error_messages = {
        'invalid': _('Enter a whole number.'),
    }


class RoundedDecimalField(fields.DecimalField):
    """
    Currency field subclass of Decimal used for rounding currencies
    to two decimal places.
    """

    def __init__(self, max_digits=None, decimal_places=2, *args, **kwargs):
        max_digits = max_digits or self.MAX_STRING_LENGTH
        super(RoundedDecimalField, self).__init__(
            max_digits=max_digits,
            decimal_places=decimal_places,
            *args, **kwargs
        )

    def to_internal_value(self, data):
        return self.quantize(super(RoundedDecimalField, self).to_internal_value(data))

    def validate_precision(self, data):
        return data


__all__ = [name for name, value in locals().items()
           if inspect.isclass(value) and issubclass(value, fields.Field)]
